#ifndef _Smart_pointers
#define _Smart_pointers

#include <cstddef> // 'size_t'

#include <list>
#include <new> // Standard exception types like 'std::bad_alloc'
#include <string>

class Elephantidae { // 'Elephant' and 'Mammoth' as subtypes...
private:
	std::list<std::shared_ptr<Elephantidae> > _family_members;
protected:
	virtual ~Elephantidae() = default; // This makes 'Elephantidae' polymorphic to be used with 'dynamic_cast' *AND* 'dynamic_pointer_cast'
public:
	void push_family_member(const std::shared_ptr<Elephantidae>&);
	std::shared_ptr<Elephantidae> peek_family_member() const;
};

class Elephant : public Elephantidae {
	static void* _Preallocated_memory;
	short _weight;
public:
	void* operator new(size_t) noexcept(false);
	void operator delete(void*) noexcept(true);
	//    void* operator new[](size_t) noexcept(false); 
	//    void operator delete[](void*) noexcept(true);
	bool operator<(const Elephant&) const;
public:
	std::string sub_species() const;
private:
	static const std::string _Loxodonta_africana;
	static const std::string _Elephas_maximus;
};

class African_elephant : private Elephant {
};

class Asian_elephant : private Elephant {
};

class Mammoth : public Elephantidae {
private:
	const Mammoth* _best_friend = nullptr;
public:
	const std::string name;

	inline Mammoth(const std::string& name = "") : name(name) {
	}; // 'inline' is for compactness only!

	inline void best_friend(const Mammoth* best_friend) {
		_best_friend = best_friend;
	}; // 'inline' is for compactness only!
};

#endif
